//============================================================================//
//
//  Software Name : MEGADOCK Version 2.1
//
//  Class Name : Control
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "control.h"

//============================================================================//
void Control::initialize(int argc,char *argv[])
//============================================================================//
{
  vector<Receptor>	_receptor;
  vector<Ligand>	_ligand;
  vector<HeaderDB>	_header;
  int			ngrid;
  vector<int>		ngrid_table;
  time_t 		ct1, ct2;
  const int		nproc	= _parallel->nproc();
  const int		myid	= _parallel->myid();

  time(&ct1);
 
  // Parameter
  _parameter = new Parameter(_parallel);
  _parameter->initialize(argc,argv);
  const int istep = _parameter->_Cal_step;

  if( !_parameter->_Num_fft_flag ) {
    gridtable(ngrid,ngrid_table);
  }

  // Receptor
  if( istep == 0 || istep == 2 ) {
    for( int i = 0 ; i < _parameter->_Num_rec ; i++ ) {
      Receptor ireceptor(_parameter->_RecPDB_file[i]);
      ireceptor.initialize(_parameter,i,nproc,myid);

      if( !_parameter->_Num_fft_flag ) {
	autogridr(ireceptor,ngrid,ngrid_table);
      }
      else {
	checkgridr(ireceptor);
      }

      _receptor.push_back(ireceptor);
    }
  }

  // Ligand
  if( istep == 0 || istep == 1 ) {
    for( int i = 0 ; i < _parameter->_Num_lig ; i++ ) {
      Ligand iligand(_parameter->_LigPDB_file[i]);
      iligand.initialize(_parameter,i,nproc,myid);

      if( !_parameter->_Num_fft_flag ) {
	autogridl(iligand,ngrid,ngrid_table);
      }
      else {
	checkgridl(iligand);
      }

      _ligand.push_back(iligand);
    }
  }
  else {
    for( int i = 0 ; i < _parameter->_Num_lig ; i++ ) {
      HeaderDB iheader(_parameter);

      if( (i%nproc) == myid ) {
	iheader.fft_read(i);
      }
#ifdef PARALLEL
      iheader.fft_comm(i,nproc,myid);
#endif

      _header.push_back(iheader);
      _parameter->_LigPDB_file.push_back(iheader.pdb_file());
    }
  }

  // Docking
  _docking = new Docking(_cputime,_parallel,_parameter,_receptor,_ligand,
			 _header);
  _docking->initialize();

  time(&ct2);
  _cputime->control_initialize += difftime(ct2,ct1);

  return;
}

//============================================================================//
void Control::gridtable(int &ngrid,vector<int> &ngrid_table)
//============================================================================//
{
  int grid_table[] = {
2,
3,
4,
5,
6,
7,
8,
9,
10,
11,
12,
14,
15,
16,
18,
20,
21,
22,
24,
25,
27,
28,
30,
32,
33,
35,
36,
40,
42,
44,
45,
48,
49,
50,
54,
55,
56,
60,
63,
64,
66,
70,
72,
75,
77,
80,
81,
84,
88,
90,
96,
98,
99,
100,
105,
108,
110,
112,
120,
121,
125,
126,
128,
132,
135,
140,
144,
147,
150,
154,
160,
162,
165,
168,
175,
176,
180,
189,
192,
196,
198,
200,
210,
216,
220,
224,
225,
231,
240,
242,
243,
245,
250,
252,
256,
264,
270,
275,
280,
288,
294,
297,
300,
308,
315,
320,
324,
330,
336,
343,
350,
352,
360,
363,
375,
378,
384,
385,
392,
396,
400,
405,
420,
432,
440,
441,
448,
450,
462,
480,
484,
486,
490,
495,
500,
504,
512,
525,
528,
539,
540,
550,
560,
567,
576,
588,
594,
600,
605,
616,
625,
630,
640,
648,
660,
672,
675,
686,
693,
700,
704,
720,
726,
729,
735,
750,
756,
768,
770,
784,
792,
800,
810,
825,
840,
847,
864,
875,
880,
882,
891,
896,
900,
924,
945,
960,
968,
972,
980,
990,
1000,
1008,
1024
  };

  ngrid = sizeof(grid_table)/sizeof(grid_table[0]);
  ngrid_table.resize(ngrid);
  if( ngrid_table.size() < ngrid ) {
    cerr << "[ERROR] Out of memory. Number of FFT table lists = ("
	 << ngrid << ") for (ngrid_table) in control.cpp!!\n";
    exit(1);
  }

  for( int i = 0 ; i < ngrid ; i++ ) {
    ngrid_table[i] = grid_table[i];
  }

  return;
}

//============================================================================//
void Control::autogridr(Receptor &ireceptor,const int &ngrid,
			vector<int> &ngrid_table)
//============================================================================//
{
  int		num_grid = 1;
  float		size, size_rec = 0.0;
  const int	myid	= _parallel->myid();

  for( int i = 0 ; i < 3 ; i++ ) {
    size = ireceptor.edge(i,1) - ireceptor.edge(i,0);

    if( size > size_rec ) {
      size_rec = size;
    }
  }

  if( myid == 0 ) {
    cout << "\nReceptor = " << ireceptor.input_file() << endl;
    cout << "Receptor max size = " << size_rec << endl;
  }

  size_rec += 2.0 * _parameter->_Grid_space_rec;
  if( myid == 0 ) {
    cout << "Required voxel size = " << size_rec << endl;
  }

  num_grid = 1 + int(size_rec / _parameter->_Grid_size);

  for( int i = 0 ; i < ngrid ; i++ ) {
    if( ngrid_table[i] >= num_grid ) {
      num_grid = ngrid_table[i];
      break;
    }
  }

  if( num_grid < _parameter->_Num_fft_l/2 ) {
    num_grid = _parameter->_Num_fft_l/2;
  }

  ireceptor.num_grid(num_grid);

  if( myid == 0 ) {
    cout << "Number of grid = " << num_grid << endl;
    cout << "FFT N = " << num_grid*2 << endl;
  }

  return;
}

//============================================================================//
void Control::autogridl(Ligand &iligand,const int &ngrid,
			vector<int> &ngrid_table)
//============================================================================//
{
  int		num_grid = 1;
  float		size_lig = 0.0;
  float		x1, y1, z1, x2, y2, z2, d2;
  const int	na	= iligand.num_atoms();
  const int	myid	= _parallel->myid();

  for( int i = 0 ; i < na-1 ; i++ ) {
    x1 = iligand.coordinate(i,0);
    y1 = iligand.coordinate(i,1);
    z1 = iligand.coordinate(i,2);

    for( int j = i+1 ; j < na ; j++ ) {
      x2 = iligand.coordinate(j,0);
      y2 = iligand.coordinate(j,1);
      z2 = iligand.coordinate(j,2);

      d2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);

      if( d2 > size_lig ) {
	size_lig = d2;
      }
    }
  }

  size_lig = sqrt(size_lig);

  if( myid == 0 ) {
    cout << "\nLigand = " << iligand.input_file() << endl;
    cout << "Ligand max size = " << size_lig << endl;
  }

  size_lig += 2.0 * _parameter->_Grid_space_lig;

  if( myid == 0 ) {
    cout << "Required voxel size = " << size_lig << endl;
  }

  num_grid = 1 + int(size_lig / _parameter->_Grid_size);

  for( int i = 0 ; i < ngrid ; i++ ) {
    if( ngrid_table[i] >= num_grid ) {
      num_grid = ngrid_table[i];
      break;
    }
  }

  if( num_grid < _parameter->_Num_fft_l/2 ) {
    num_grid = _parameter->_Num_fft_l/2;
  }

  iligand.num_grid(num_grid);

  if( myid == 0 ) {
    cout << "Number of grid = " << num_grid << endl;
    cout << "FFT N = " << num_grid*2 << endl;
  }

  return;
}

//============================================================================//
void Control::checkgridr(Receptor &ireceptor)
//============================================================================//
{
  float		size, size_rec	= 0.0;
  const int	num_grid	= _parameter->_Num_grid;
  const float	search_length	= _parameter->_Grid_size * num_grid;
  const int	myid		= _parallel->myid();

  for( int i = 0 ; i < 3 ; i++ ) {
    size = ireceptor.edge(i,1) - ireceptor.edge(i,0);

    if( size > size_rec ) {
      size_rec = size;
    }
  }

  if( myid == 0 ) {
    cout << "\nReceptor max size = " << size_rec << endl;
  }

  size_rec += 2.0*_parameter->_Grid_space_rec;

  if( myid == 0 ) {
    cout << "Required voxel size = " << size_rec << endl;
  }

  if( size_rec > search_length ) {
    cerr << "[ERROR] Receptor data is too big!!\n";
    exit(1);
  }

  ireceptor.num_grid(num_grid);

  if( myid == 0 ) {
    cout << "\n(Receptor)\n";
    cout << "Number of grid = " << num_grid << endl;
    cout << "FFT N = " << num_grid*2 << endl;
    cout << "Grid size = " << _parameter->_Grid_size << endl;
  }

  return;
}

//============================================================================//
void Control::checkgridl(Ligand &iligand)
//============================================================================//
{
  float		size_lig = 0.0;
  float		x1, y1, z1, x2, y2, z2, d2;
  const int	na	= iligand.num_atoms();
  const int	num_grid	= _parameter->_Num_grid;
  const float	search_length	= _parameter->_Grid_size * num_grid;
  int		myid	= _parallel->myid();

  for( int i = 0 ; i < na-1 ; i++ ) {
    x1 = iligand.coordinate(i,0);
    y1 = iligand.coordinate(i,1);
    z1 = iligand.coordinate(i,2);

    for( int j = i+1 ; j < na ; j++ ) {
      x2 = iligand.coordinate(j,0);
      y2 = iligand.coordinate(j,1);
      z2 = iligand.coordinate(j,2);

      d2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);

      if( d2 > size_lig ) {
	size_lig = d2;
      }
    }
  }

  size_lig = sqrt(size_lig);
  if( myid == 0 ) {
    cout << "\nLigand max size = " << size_lig << endl;
  }

  size_lig += 2.0*_parameter->_Grid_space_lig;
  if( myid == 0 ) {
    cout << "Required voxel size = " << size_lig << endl;
  }

  if( size_lig > search_length ) {
    cerr << "[ERROR] Ligand data is too big!!\n";
    exit(1);
  }

  iligand.num_grid(num_grid);

  if( myid == 0 ) {
    cout << "\n(Ligand)\n";
    cout << "Number of grid = " << num_grid << endl;
    cout << "FFT N = " << num_grid*2 << endl;
    cout << "Grid size = " << _parameter->_Grid_size << endl;
  }

  return;
}

//============================================================================//
void Control::execute()
//============================================================================//
{
  const int	istep 	= _parameter->_Cal_step;
  const int	nlig 	= _parameter->_Num_lig;
  time_t 	ct1, ct2;
  const int	nproc	= _parallel->nproc();
  const int	myid	= _parallel->myid();

  if( istep != 1 ) {
    if( myid == 0 ) {
      cout << "\n---------- Start docking calculations\n";
      if(_docking->fft_nlist() > 1){
	cout << "FFT N list = (";
	for( int i = 0 ; i < _docking->fft_nlist() ; i++ ) {
	  cout << " " << _docking->fft_list(i);
	}
	cout << " )" << endl;
      }
    }
  }

  if( istep == 0 ) {
    for( int i = 0 ; i < _docking->fft_nlist() ; i++ ) {
      if( myid == 0 ) {
	cout << "Now FFT N = " << _docking->fft_list(i) << endl;
      }

      const int nlist_lig = _docking->lig_nlist(i);

      time(&ct1);
      _docking->rec_init(i);
      time(&ct2);
      _cputime->receptor_process += difftime(ct2,ct1);

      for( int j = 0 ; j < nlist_lig ; j++ ) {
	time(&ct1);
	_docking->dockz(i,j);
	time(&ct2);
	_cputime->docking_dock += difftime(ct2,ct1);

	time(&ct1);
	_docking->output(i,j);
	time(&ct2);
	_cputime->docking_output += difftime(ct2,ct1);
      }
    }
  }
  else if( istep == 1 ) {
    time(&ct1);
    for( int i = myid ; i < nlig ; i+=nproc ) {
      _docking->dockz1(i);
    }
    time(&ct2);
    _cputime->docking_dock += difftime(ct2,ct1);
  }
  else if( istep == 2 ) {
    int ic = 0;

    for( int i = 0 ; i < _docking->fft_nlist() ; i++ ) {
      if( myid == (ic%nproc) ) {
	cout << "\nNow FFT N = " << _docking->fft_list(i)
	     << " (myid=" << myid << ")" << endl;
      }

      const int nlist_lig = _docking->lig_nlist(i);

      time(&ct1);
      _docking->rec_init(i);
      time(&ct2);
      _cputime->receptor_process += difftime(ct2,ct1);

      for( int j = 0 ; j < nlist_lig ; j++ ) {
	if( myid == (ic%nproc) ) {
	  time(&ct1);
	  _docking->dock2(i,j);
	  time(&ct2);
	  _cputime->docking_dock += difftime(ct2,ct1);

	  time(&ct1);
	  _docking->output(i,j);
	  time(&ct2);
	  _cputime->docking_output += difftime(ct2,ct1);
	}

	ic++;
      }
    }
  }

  _docking->delete_array();
  _docking->delete_class();
  delete _docking;
  delete _parameter;

  return;
}
