//============================================================================//
//
//  Software Name : MEGADOCK Version 2.1
//
//  Class Name : FFTProcess
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//
//
//  Reference :
//    Katchalski-Katzir E, Shariv I, Eisenstein M, Friesem AA, Aflalo C,
//    Vakser IA, "Molecular surface recognition: Determination of geometric
//    fit between proteins and their ligands by correlation techniques",
//    Proc Natl Acad Sci USA, 89,2195-2199(1992).
//
//============================================================================//

#include "fft_process.h"

//============================================================================//
bool ascending_struct(const SortScore &s1,const SortScore &s2)	// for sort
//============================================================================//
{
  return s1.score > s2.score;
}

//============================================================================//
void FFTProcess::alloc_array(const int &maxlist,const int &maxfft,
			     const int &maxarray)
//============================================================================//
{
  const int	istep	= _parameter->_Cal_step;
  const int 	ns 	= _parameter->_Num_sort;
  const int 	nz 	= _parameter->_NZangle;
  const int 	nf3 	= maxfft * maxfft * maxfft;
  const int 	no 	= _parameter->_Num_output;
  const int	nproc	= _parallel->nproc();
  const int	nproc2	= _parallel->nproc2();
  int	nt;

  if( istep == 0 ) {
    nt = ( (nz-1)/nproc ) * ns + ns;
  }
  else {
    nt = nz * ns;
  }

  if( no > nt ) nt = no;

  //  if( nproc2 == 1 ) alloc_fft(maxfft);
  alloc_fft(maxfft);

  if( istep != 1 ) {
    _Select.resize(nproc2);
    for( int i = 0 ; i < nproc2 ; i++ ) {
      _Select[i].resize(ns);
    }

    _Top.resize(maxlist);
    for( int i = 0 ; i < maxlist ; i++ ) {
      _Top[i].resize(nt);
    }
  }

  //---------- memory allocation for _Rotation
  _Rotation = new int[nproc2];

  if( istep != 1 ) {
    //---------- memory allocation for _FFT_rec_r
    _FFT_rec_r = new float[maxarray];
    if( !_FFT_rec_r ) {
      cerr << "[ERROR] Out of memory. Number of listed receptors = ("
	   << maxarray << ") for (_FFT_rec_r) in fft_process.cpp!!\n";
      exit(1);
    }

    //---------- memory allocation for _FFT_rec_i
    _FFT_rec_i = new float[maxarray];
    if( !_FFT_rec_i ) {
      cerr << "[ERROR] Out of memory. Number of listed receptors = ("
	   << maxarray << ") for (_FFT_rec_i) in fft_process.cpp!!\n";
      exit(1);
    }
  }

  //---------- memory allocation for _FFT_lig_r
  _FFT_lig_r = new float*[nproc2];
  for( int i = 0 ; i < nproc2 ; i++ ) {
    _FFT_lig_r[i] = new float[nf3];
    if( !_FFT_lig_r[i] ) {
      cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
	   << nf3*nproc2 << ") for (_FFT_lig_r) in fft_process.cpp!!\n";
      exit(1);
    }
  }

  //---------- memory allocation for _FFT_lig_i
  _FFT_lig_i = new float*[nproc2];
  for( int i = 0 ; i < nproc2 ; i++ ) {
    _FFT_lig_i[i] = new float[nf3];
    if( !_FFT_lig_i[i] ) {
      cerr << "[ERROR] Out of memory. Number of FFT N*N*N = ("
	   << nf3*nproc2 << ") for (_FFT_lig_i) in fft_process.cpp!!\n";
      exit(1);
    }
  }

  return;
}

//============================================================================//
void FFTProcess::alloc_fft(const int &maxfft)
//============================================================================//
{
  const int	nf3 = maxfft * maxfft * maxfft;
  const int	nproc2	= _parallel->nproc2();

  _FFTWin  = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);
  _FFTWout = (fftwf_complex*) fftwf_malloc(sizeof(fftwf_complex)*nf3*nproc2);

  return;
}

//============================================================================//
void FFTProcess::delete_fft()
//============================================================================//
{

  fftwf_free(_FFTWin); 
  fftwf_free(_FFTWout); 

  return;
}

//============================================================================//
void FFTProcess::delete_array()
//============================================================================//
{
  const int	istep	= _parameter->_Cal_step;
  const int	nproc2	= _parallel->nproc2();

  if( istep != 1 ) {
    vector< vector<SortScore> > tmp1, tmp2;

    _Select.swap(tmp1);
    _Top.swap(tmp2);
  }

  //  if( nproc2 == 1 ) delete_fft();
  delete_fft();

  delete [] _Rotation;

  if( istep != 1 ) {
    delete [] _FFT_rec_r;
    delete [] _FFT_rec_i;
  }

  delete [] _FFT_lig_r;
  delete [] _FFT_lig_i;

  return;
}

//============================================================================//
void FFTProcess::receptor_fft(const int &ilist_rec,float *grid_r,float *grid_i)
//============================================================================//
{
  const int	num_grid= _Num_fft / 2;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  const int	ndata	= ( _Num_fft - num_grid ) / 2;
  const int	ir	= ilist_rec * nf3;
  const float	theta	= -2.0 * PI / _Num_fft;
  const int	nproc2	= _parallel->nproc2();

  //  if( nproc2 > 1 ) alloc_fft(_Num_fft);

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[i][0] = 0.0;
    _FFTWin[i][1] = 0.0;
  }

  for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
    const int ic = _Num_fft*_Num_fft*(i+ndata);

    for( int j = 0 ; j < num_grid ; j++ ) {
      const int jc = ic + _Num_fft*(j+ndata);

      for( int k = 0 ; k < num_grid ; k++ ) {
	_FFTWin[jc+k+ndata][0] = grid_r[m  ];
	_FFTWin[jc+k+ndata][1] = grid_i[m++];
      }
    }
  }

  fft3d(theta,0);

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFT_rec_r[ir+i] = _FFTWout[i][0];
    _FFT_rec_i[ir+i] = _FFTWout[i][1];
  }

  //  if( nproc2 > 1 ) delete_fft();

  return;
}

//============================================================================//
void FFTProcess::ligand_fft(float *grid_r,float *grid_i,int myid2)
//============================================================================//
{
  const int 	num_grid = _Num_fft / 2;
  const int	ndata	= ( _Num_fft - num_grid ) / 2;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  const float	theta	= -2.0 * PI / _Num_fft;
  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[nf3*myid2+i][0]	= 0.0;
    _FFTWin[nf3*myid2+i][1]	= 0.0;
  }

  for( int i = 0, m = 0 ; i < num_grid ; i++ ) {
    const int ic = _Num_fft*_Num_fft*(i+ndata);

    for( int j = 0 ; j < num_grid ; j++ ) {
      int jc = ic + _Num_fft*(j+ndata);

      for( int k = 0 ; k < num_grid ; k++ ) {
	_FFTWin[nf3*myid2+jc+k+ndata][0]	= grid_r[m  ];
	_FFTWin[nf3*myid2+jc+k+ndata][1]	= grid_i[m++];
      }
    }
  }

  fft3d(theta,myid2);

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFT_lig_r[myid2][i] = _FFTWout[nf3*myid2+i][0];
    _FFT_lig_i[myid2][i] = _FFTWout[nf3*myid2+i][1];
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_ligand_fft += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::fft_score(const int &ilist_rec,int myid2)
//============================================================================//
{
  const int	nf3 	= _Num_fft * _Num_fft * _Num_fft;
  const int	ir	= ilist_rec * nf3;
  const float	theta 	= 2.0 * PI / _Num_fft;
  const int	nproc2	= _parallel->nproc2();
  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    _FFTWin[nf3*myid2+i][0] = _FFT_rec_r[ir+i]*_FFT_lig_r[myid2][i]
      + _FFT_rec_i[ir+i]*_FFT_lig_i[myid2][i];
    _FFTWin[nf3*myid2+i][1] = _FFT_rec_r[ir+i]*_FFT_lig_i[myid2][i]
      - _FFT_rec_i[ir+i]*_FFT_lig_r[myid2][i];
  }

  fft3d(theta,myid2);

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_fft_inverse += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::fft3d(const float &theta,int myid2)
//============================================================================//
{
  const int	n   = _Num_fft;
  const int	nf3 = n*n*n;
  fftwf_plan	plan;

#pragma omp critical
  {
    if( theta < 0.0 ) {
      plan=fftwf_plan_dft_3d(n,n,n,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],
			     FFTW_FORWARD,FFTW_ESTIMATE); 
    }
    else {
      plan=fftwf_plan_dft_3d(n,n,n,&_FFTWin[nf3*myid2],&_FFTWout[nf3*myid2],
			     FFTW_BACKWARD,FFTW_ESTIMATE); 
    }
  }

  fftwf_execute(plan); 

#pragma omp critical
  {
    fftwf_destroy_plan(plan);
  }

  return;
}

//============================================================================//
void FFTProcess::score_sort(const int &ilist_rec,int myid2)
//============================================================================//
{
  const int	ns	= _parameter->_Num_sort;
  const int	nf2	= _Num_fft * _Num_fft;
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  time_t	ct1, ct2;
  const int	istep 	= _parameter->_Cal_step;
  const int	nproc	= _parallel->nproc();
  const int	nproc2	= _parallel->nproc2();

  if( myid2 == 0 ) {
    time(&ct1);
  }

  for( int i = 0 ; i < ns ; i++ ) {
    _Select[myid2][i].score = 0.0;
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    const float raw = _FFTWout[nf3*myid2+i][0] / nf3;

    for( int j = 0 ; j < ns ; j++ ) {
      if( raw > _Select[myid2][j].score ) {
	for( int k = ns-1 ; k > j ; k-- ) {
	  _Select[myid2][k] = _Select[myid2][k-1];
	}
	_Select[myid2][j].score	= raw;
	_Select[myid2][j].index[1]	= i / nf2;
	_Select[myid2][j].index[2]	= (i / _Num_fft) % _Num_fft;
	_Select[myid2][j].index[3]	= i % _Num_fft;
	break;
      }
    }
  }

  for( int i = 0 ; i < ns ; i++ ) {
    _Select[myid2][i].index[0] = _Rotation[myid2];
  }

  if( istep == 0 ) {
    for( int i = 0 ; i < ns ; i++ ) {
      _Top[ilist_rec][(_Rotation[myid2]/nproc)*ns+i] = _Select[myid2][i];
    }
  }
  else {
    for( int i = 0 ; i < ns ; i++ ) {
      _Top[ilist_rec][_Rotation[myid2]*ns+i] = _Select[myid2][i];
    }
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->fftprocess_score_sort += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void FFTProcess::fftoutl(ofstream &fftlig)
//============================================================================//
{
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;

#ifdef CHARACTER
  for( int i = 0 ; i < nf3 ; i++ ) {
    fftlig << _FFT_lig_r[0][i] << " ";
  }

  fftlig << endl;

  for( int i = 0 ; i < nf3 ; i++ ) {
    fftlig << _FFT_lig_i[0][i] << " ";
  }

  fftlig << endl;
#else
  fftlig.write((char*)_FFT_lig_r[0],sizeof(float)*nf3);
  fftlig.write((char*)_FFT_lig_i[0],sizeof(float)*nf3);
#endif

  return;
}

//============================================================================//
void FFTProcess::fftinl(ifstream &fftlig,const int &im)
//============================================================================//
{
  const int	nf3	= _Num_fft * _Num_fft * _Num_fft;
  time_t	ct1, ct2;

  time(&ct1);

#ifdef CHARACTER
  for( int i = 0 ; i < nf3 ; i++ ) {
    fftlig >> _FFT_lig_r[im][i];
  }

  for( int i = 0 ; i < nf3 ; i++ ) {
    fftlig >> _FFT_lig_i[im][i];
  }
#else
  fftlig.read((char*)_FFT_lig_r[im],sizeof(float)*nf3);
  fftlig.read((char*)_FFT_lig_i[im],sizeof(float)*nf3);
#endif

  time(&ct2);
  _cputime->fftprocess_fftinl += difftime(ct2,ct1);

  return;
}

//============================================================================//
void FFTProcess::sort_index(const int &nlist_rec,float *fwork,int *iwork)
//============================================================================//
{
  const int 	no 	= _parameter->_Num_output;
  const int 	nt 	= (_Top[0].size() < no) ? _Top[0].size() : no;
  const int	istep 	= _parameter->_Cal_step;
  const int	nproc	= _parallel->nproc();

  for( int i = 0 ; i < nlist_rec ; i++ ) {
    partial_sort(_Top[i].begin(),_Top[i].begin()+nt,_Top[i].end(),
		 ascending_struct);
  }

  if( istep == 2 ) {
    return;
  }

#ifdef PARALLEL
  MPI_Status    status;
  float		*top_score   = fwork;
  float 	*tmp_score   = &fwork[no];
  float 	*stock_score = &fwork[no*2];
  int		*top_index   = iwork;
  int		*tmp_index   = &iwork[no*4];
  int		*stock_index = &iwork[no*8];
  const int	myid	= _parallel->myid();

  for( int ir = 0 ; ir < nlist_rec ; ir++ ) {
    const int ip = ir % nproc;

    if( myid == ip ) {
      for( int i = 0, i1 = 1 ; i < nproc ; i++ ) {
	if( i == ip ) continue;

	MPI_Recv(top_score,nt,MPI_FLOAT,i,100,MPI_COMM_WORLD,&status);
	MPI_Recv(top_index,nt*4,MPI_INT,i,110,MPI_COMM_WORLD,&status);

	const int nt0 = (no < i1*nt) ? no : i1*nt;
	i1++;
	const int nt2 = (no < i1*nt) ? no : i1*nt;

	for( int j = 0, j1 = 0, j2 = 0 ; j < nt2 ; j++ ) {
	  if( j1 >= nt0 ) {
	    for( int k = j ; k < nt ; k++ ) {
	      stock_score[j] = _Top[ir][j2].score;

	      for( int k = 0 ; k < 4 ; k++ ) {
		stock_index[4*j+k] = _Top[ir][j2].index[k];
	      }

	      j2++;
	    }

	    break;
	  }
	  else if( j2 >= nt ) {
	    for( int k = j ; k < nt2 ; k++ ) {
	      stock_score[j] = _Top[ir][j1].score;

	      for( int k = 0 ; k < 4 ; k++ ) {
		stock_index[4*j+k] = _Top[ir][j1].index[k];
	      }

	      j1++;
	    }

	    break;
	  }

	  if( _Top[ir][j1].score > top_score[j2] ) {
	    stock_score[j] = _Top[ir][j1].score;

	    for( int k = 0 ; k < 4 ; k++ ) {
	      stock_index[4*j+k] = _Top[ir][j1].index[k];
	    }

	    j1++;
	  }
	  else {
	    stock_score[j] = top_score[j2];
	    
	    for( int k = 0 ; k < 4 ; k++ ) {
	      stock_index[4*j+k] = top_index[4*j2+k];
	    }

	    j2++;
	  }
	}

	for( int j = 0 ; j < nt2 ; j++ ) {
	  _Top[ir][j].score = stock_score[j];

	  for( int k = 0 ; k < 4 ; k++ ) {
	    _Top[ir][j].index[k] = stock_index[4*j+k];
	  }
	}
      }
    }
    else {
      for( int j = 0 ; j < nt ; j++ ) {
	tmp_score[j] = _Top[ir][j].score;
      }

      for( int j = 0 ; j < nt ; j++ ) {
	for( int k = 0 ; k < 4 ; k++ ) {
	  tmp_index[j*4+k] = _Top[ir][j].index[k];
	}
      }

      MPI_Send(tmp_score,nt,MPI_FLOAT,ip,100,MPI_COMM_WORLD);
      MPI_Send(tmp_index,nt*4,MPI_INT,ip,110,MPI_COMM_WORLD);
    }
  }
#endif

  return;
}

//============================================================================//
void FFTProcess::top_score_clean(const int &nlist_rec)
//============================================================================//
{
  const int 	ns 	= _parameter->_Num_sort;
  const int 	nz 	= _parameter->_NZangle;
  const int 	no 	= _parameter->_Num_output;
  const int	istep 	= _parameter->_Cal_step;
  const int	nproc	= _parallel->nproc();
  int	nt;

  if( istep == 0 ) {
    nt = ( (nz-1)/nproc ) * ns + ns;
  }
  else {
    nt = nz * ns;
  }

  if( no > nt ) nt = no;

  for( int i = 0 ; i < nlist_rec ; i++ ) {
    for( int j = 0 ; j < nt ; j++ ) {
      _Top[i][j].score = 0.0;
    }
  }

  return;
}
