//============================================================================//
//
//  Software Name : MEGADOCK Version 2.1
//
//  Class Name : Receptor
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "receptor.h"

//============================================================================//
void Receptor::katchalski(const int &score_func,const int &num_grid,
			  float *grid_coord,float *mol_coord,float *fwork)
//============================================================================//
{
  float	rec_inside;

  if( score_func == 1 ) {		// for MEGADOCK parameter
    rec_inside = -2.0;
  }
  else {				// for FTDock
    rec_inside = -1.5;			// K-K original parameter
  }

  katchalski_base(num_grid,rec_inside,grid_coord,mol_coord,fwork);

  return;
}

//============================================================================//
void Receptor::gsc(const int &num_grid,float *grid_coord,float *mol_coord,
		   int *iwork,float *fwork)
//============================================================================//
{
  const	float	rho = -1.0;			// MEGADOCK parameter

  int		*nearesta;
  float		*grid_r, *grid_i;
  float		*rjudge2, *radius_core2, *radius_surf2;
  float		*distance3;
  float		*xd, *yd, *zd;

  const int	na  = num_atoms();
  const int	nag = num_atoms() * num_grid;
  const int	ng2 = num_grid * num_grid;
  const int	ng3 = ng2 * num_grid;

  nearesta 	= iwork;
  grid_r	= fwork;
  grid_i	= &fwork[ng3];
  rjudge2 	= &fwork[ng3*2];
  radius_core2  = &fwork[ng3*2+na];
  radius_surf2  = &fwork[ng3*2+na*2];
  distance3     = &fwork[ng3*2+na*3];
  xd		= &fwork[ng3*3+na*3];
  yd		= &fwork[ng3*3+na*3+nag];
  zd		= &fwork[ng3*3+na*3+nag*2];

  precalc(num_grid,nearesta,rjudge2,radius_core2,radius_surf2);
  voxel_init(num_grid,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);

  for( int l = 0 ; l < na ; l++ ) {
    const int lc = num_grid * l;

    for( int i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > rjudge2[l] ) continue;
      const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > rjudge2[l] ) continue;
	const float	d2 = xd[lc+i] + yd[lc+j];
	const int	jb = ib + num_grid * j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const float d3 = d2 + zd[lc+k];
	  if( d3 > rjudge2[l] ) continue;
	  const int kb = jb + k;

	  if( d3 < distance3[kb] ) {
	    distance3[kb] = d3;
	    nearesta[kb] = l;
	  }
	}
      }
    }
  }

  for( int i = 0 ; i < ng3 ; i++ ) {
    if( nearesta[i] == -99 ) continue;

    if( distance3[i] <= rjudge2[nearesta[i]] ) {
      if( distance3[i] <= radius_core2[nearesta[i]] ) {
	grid_i[i] = 9.0 * rho;
      }
      else {
	grid_r[i] = 1;
      }
    }
  }

  const int ng1 = num_grid;

  for( int i = 1 ; i < num_grid-1 ; i++ ) {
    const int ib = ng2 * i;

    for( int j = 1 ; j < num_grid-1 ; j++ ) {
      const int jb = ib + num_grid * j;

      for( int k = 1 ; k < num_grid-1 ; k++ ) {
	const int kb = jb + k;

	if( grid_i[kb] == 9.0*rho ) {
	  if( grid_i[kb-ng2] == 0.0 || grid_i[kb+ng2] == 0.0 ||
	      grid_i[kb-ng1] == 0.0 || grid_i[kb+ng1] == 0.0 ||
	      grid_i[kb-1  ] == 0.0 || grid_i[kb+1  ] == 0.0 ) {
	    if( distance3[kb] > radius_surf2[nearesta[kb]] ) {
	      grid_i[kb] = -9999.0;
	    }
	  }
	}
      }
    }
  }

  for( int i = ng2 ; i < ng3-1 ; i++ ) {
    if( grid_i[i] == -9999.0 ) {
      grid_r[i] = 1;
      grid_i[i] = 0.0;
    }
  }

  return;
}

//============================================================================//
void Receptor::psc(const int &num_grid,float *grid_coord,float *mol_coord,
		   int *iwork,float *fwork)
//============================================================================//
{
  const	float	rho = -3.0;			// MEGADOCK parameter

  int		*nearesta;
  int		*surf_atom;
  int		*counter;
  float		*grid_r, *grid_i;
  float		*rjudge2, *radius_core2, *radius_surf2;
  float		*distance3;
  float		*xd, *yd, *zd;

  const int	na = num_atoms();
  const int	nag = num_atoms() * num_grid;
  const int	ng2 = num_grid * num_grid;
  const int	ng3 = ng2 * num_grid;

  nearesta 	= iwork;
  surf_atom	= &iwork[ng3];	
  counter	= &iwork[ng3+na];	
  grid_r	= fwork;
  grid_i	= &fwork[ng3];
  rjudge2 	= &fwork[ng3*2];
  radius_core2  = &fwork[ng3*2+na];
  radius_surf2  = &fwork[ng3*2+na*2];
  distance3     = &fwork[ng3*2+na*3];
  xd		= &fwork[ng3*3+na*3];
  yd		= &fwork[ng3*3+na*3+nag];
  zd		= &fwork[ng3*3+na*3+nag*2];

  precalc(num_grid,nearesta,rjudge2,radius_core2,radius_surf2);
  voxel_init(num_grid,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);

  for( int i = 0 ; i < ng3 ; i++ ) {
    counter[i] = 0;
  }

  for( int l = 0 ; l < na ; l++ ) {
    const int lc = num_grid * l;

    for( int i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > rjudge2[l] ) continue;
      const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > rjudge2[l] ) continue;
	const float	d2 = xd[lc+i] + yd[lc+j];
	const int	jb = ib + num_grid*j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const float d3 = d2 + zd[lc+k];
	  if( d3 > rjudge2[l] ) continue;
	  const int kb = jb + k;

	  if( d3 < distance3[kb] ) {
	    distance3[kb] = d3;
	    nearesta[kb] = l;
	  }

	  counter[kb] ++;
	}
      }
    }
  }

  for( int i = 0 ; i < ng3 ; i++ ) {
    if( nearesta[i] == -99 ) continue;

    if( distance3[i] <= rjudge2[nearesta[i]] ) {
      if( distance3[i] <= radius_core2[nearesta[i]] ) {
	grid_i[i] = 9.0 * rho;
      }
      else {
	grid_r[i] = (float)counter[i];
      }
    }
  }

  const int ng1 = num_grid;
  int	nsurf = 0;

  for( int i = 1 ; i < num_grid-1 ; i++ ) {
    const int ib = ng2 * i;

    for( int j = 1 ; j < num_grid-1 ; j++ ) {
      const int jb = ib + num_grid * j;

      for( int k = 1 ; k < num_grid-1 ; k++ ) {
	const int kb = jb + k;

	if( grid_i[kb] == 9.0*rho ) {
	  if( grid_i[kb-ng2] == 0.0 || grid_i[kb+ng2] == 0.0 ||
	      grid_i[kb-ng1] == 0.0 || grid_i[kb+ng1] == 0.0 ||
	      grid_i[kb-1  ] == 0.0 || grid_i[kb+1  ] == 0.0 ) {
	    if( distance3[kb] > radius_surf2[nearesta[kb]] ) {
	      grid_i[kb] = -9999.0;
	    }
	    else {
	      int l;

	      for( l = 0 ; l < nsurf ; l++ ) {
		if( surf_atom[l] == nearesta[kb] ) {
		  break;
		}
	      }

	      if( l >= nsurf ) {
		surf_atom[nsurf++] = nearesta[kb];
	      }
	    }
	  }
	}
      }
    }
  }

  for( int i = ng2 ; i < ng3-1 ; i++ ) {
    if( grid_i[i] == -9999.0 ) {
      grid_r[i] = (float)counter[i];
      grid_i[i] = 0.0;
    }
  }

  for( int l = 0 ; l < nsurf ; l++ ) {
    const int lc = num_grid * surf_atom[l];

    for( int i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > radius_surf2[surf_atom[l]] ) continue;
      const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > radius_surf2[surf_atom[l]] ) continue;
	const float	d2 = xd[lc+i] + yd[lc+j];
	const int	jb = ib + num_grid * j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const float d3 = d2 + zd[lc+k];

	  if( d3 <= radius_surf2[surf_atom[l]] ) {
	    grid_r[jb+k] = 0.0;
	    grid_i[jb+k] = 3.0 * rho;
	  }
	}
      }
    }
  }

  return;
}

//============================================================================//
void Receptor::rpsc(const int &num_grid,float *grid_coord,float *mol_coord,
		    int *iwork,float *fwork)
//============================================================================//
{
  const	float	rho = -3.0;			// MEGADOCK parameter

  int		*nearesta;
  int		*surf_atom;
  int		*counter;
  float		*grid_r, *grid_i;
  float		*rjudge2, *radius_core2, *radius_surf2;
  float		*distance3;
  float		*xd, *yd, *zd;

  const int	na  = num_atoms();
  const int	nag = num_atoms() * num_grid;
  const int	ng2 = num_grid * num_grid;
  const int	ng3 = num_grid * num_grid * num_grid;

  nearesta 	= iwork;
  surf_atom	= &iwork[ng3];	
  counter	= &iwork[ng3+na];	
  grid_r	= fwork;
  grid_i	= &fwork[ng3];
  rjudge2 	= &fwork[ng3*2];
  radius_core2  = &fwork[ng3*2+na];
  radius_surf2  = &fwork[ng3*2+na*2];
  distance3     = &fwork[ng3*2+na*3];
  xd		= &fwork[ng3*3+na*3];
  yd		= &fwork[ng3*3+na*3+nag];
  zd		= &fwork[ng3*3+na*3+nag*2];

  precalc(num_grid,nearesta,rjudge2,radius_core2,radius_surf2);
  voxel_init(num_grid,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);

  for( int i = 0 ; i < ng3 ; i++ ) {
    counter[i] = 0;
  }

  int 	i, j, k, lc, kb;
  float d2, d3;

#ifdef _OPENMP
#pragma omp parallel for private(i,j,k,lc,d2,d3,kb)
#endif
  for( int l = 0 ; l < na ; l++ ) {
    lc = num_grid * l;

    for( i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > rjudge2[l] ) continue;
      //      const int ib = ng2 * i;

      for( j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > rjudge2[l] ) continue;
	d2 = xd[lc+i] + yd[lc+j];
	//	const int	jb = ib + num_grid*j;

	for( k = 0 ; k < num_grid ; k++ ) {
	  d3 = d2 + zd[lc+k];
	  if( d3 > rjudge2[l] ) continue;
	  //	  const int kb = jb + k;
	  kb = ng2*i + num_grid*j + k;

#pragma omp critical
	  {
	    if( d3 < distance3[kb] ) {
	      distance3[kb] = d3;
	      nearesta[kb] = l;
	    }

	    counter[kb] ++;
	  }
	}
      }
    }
  }

  for( int i = 0 ; i < ng3 ; i++ ) {
    if( nearesta[i] == -99 ) continue;

    if( distance3[i] <= rjudge2[nearesta[i]] ) {
      if( distance3[i] <= radius_core2[nearesta[i]] ) {
	grid_r[i] = 9.0 * rho;
      }
      else {
	grid_r[i] = (float)counter[i];
      }
    }
  }

  const int ng1 = num_grid;
  int	nsurf = 0;

  for( int i = 1 ; i < num_grid-1 ; i++ ) {
    const int ib = ng2 * i;

    for( int j = 1 ; j < num_grid-1 ; j++ ) {
      const int jb = ib + num_grid * j;

      for( int k = 1 ; k < num_grid-1 ; k++ ) {
	const int kb = jb + k;

	if( grid_r[kb] == 9.0*rho ) {
	  /*
	  if( grid_r[kb-ng2] == (float)counter[kb-ng2] ||
	      grid_r[kb+ng2] == (float)counter[kb+ng2] ||
	      grid_r[kb-ng1] == (float)counter[kb-ng1] ||
	      grid_r[kb+ng1] == (float)counter[kb+ng1] ||
	      grid_r[kb-1  ] == (float)counter[kb-1  ] ||
	      grid_r[kb+1  ] == (float)counter[kb+1  ] ) {
	  */
// Bug start
	  if( grid_r[kb-ng2] == 0.0 ||
	      grid_r[kb+ng2] == 0.0 ||
	      grid_r[kb-ng1] == 0.0 ||
	      grid_r[kb+ng1] == 0.0 ||
	      grid_r[kb-1  ] == 0.0 ||
	      grid_r[kb+1  ] == 0.0 ) {
// Bug end
	    if( distance3[kb] > radius_surf2[nearesta[kb]] ) {
	      grid_r[kb] = -9999.0;
	    }
	    else {
	      int l;

	      for( l = 0 ; l < nsurf ; l++ ) {
		if( surf_atom[l] == nearesta[kb] ) {
		  break;
		}
	      }

	      if( l >= nsurf ) {
		surf_atom[nsurf++] = nearesta[kb];
	      }
	    }
	  }
	}
      }
    }
  }

  for( int i = ng2 ; i < ng3-ng2 ; i++ ) {
    if( grid_r[i] == -9999.0 ) {
      grid_r[i] = (float)counter[i];
    }
  }

  for( int l = 0 ; l < nsurf ; l++ ) {
    const int lc = num_grid * surf_atom[l];

    for( int i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > radius_surf2[surf_atom[l]] ) continue;
      const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > radius_surf2[surf_atom[l]] ) continue;
	const float	d2 = xd[lc+i] + yd[lc+j];
	const int	jb = ib + num_grid * j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const float d3 = d2 + zd[lc+k];

	  if( d3 <= radius_surf2[surf_atom[l]] ) {
	    grid_r[jb+k] = 3.0 * rho;
	  }
	}
      }
    }
  }

  return;
}

//============================================================================//
void Receptor::electro(const float &beta,const float &eratio,
		       const int &num_grid,float *grid_coord,float *mol_coord,
		       int *iwork,float *fwork, float grid_size)
//============================================================================//
{
  const float	ftr1	= 6.0;			// FTDock parameter
  const float	ftr2	= 8.0;			// FTDock parameter
  const float	er1	= 4.0;			// FTDock parameter
  const float	er2	= 38.0;			// FTDock parameter
  const float	er3	= -224.0;		// FTDock parameter
  const float	er4	= 80.0;			// FTDock parameter

  float		*grid_r;
  float		*grid_i;
  const float	*xd, *yd, *zd;

  const int	na = num_atoms();;
  const int	nag = num_atoms() * num_grid;
  const int	ng2 = num_grid * num_grid;
  const int	ng3 = num_grid * num_grid * num_grid;

  float 	er, d2, d3;

  grid_r	= fwork;
  grid_i	= &fwork[ng3];
  xd		= &fwork[ng3*3+na*3];
  yd		= &fwork[ng3*3+na*3+nag];
  zd		= &fwork[ng3*3+na*3+nag*2];

  //#ifdef _OPENMP
  //#pragma omp parallel for
  //#endif
  for( int l = 0 ; l < na ; l++ ) {
    if( fabs(_Charge[l]) <= EPS ) continue;
    const int lc = num_grid * l;

    for( int i = 0 ; i < num_grid ; i++ ) {
      //const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	d2 = xd[lc+i] + yd[lc+j];
	//const int jb = ib + num_grid * j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const int kb = ng2*i + num_grid*j + k;
	  if( grid_r[kb] < 0.0 ) continue;

	  d3 = d2 + zd[lc+k];
	  d3 = sqrt(d3);

	  if( d3 <= ftr1 ) {
	    er = er1;
	  }
	  else if( d3 < ftr2 ) {
	    er = er2*d3 + er3;
	  }
	  else {
	    er = er4;
	  }

	  grid_i[kb] += eratio * beta * _Charge[l] / er / d3;
	}
      }
    }
  }

  return;
}

//============================================================================//
void Receptor::precalc(const int &num_grid,int *nearesta,float *rjudge2,
		       float *radius_core2,float *radius_surf2)
//============================================================================//
{
  const float	D = 3.6;			// ZDOCK parameter
  const	float	rcore2 = 1.5;			// ZDOCK parameter
  const	float	rsurf2 = 0.8;			// ZDOCK parameter

  const int	na  = num_atoms();
  const int	ng3 = num_grid * num_grid * num_grid;

  for( int i = 0 ; i < na ; i++ ) {
    rjudge2[i] 	    = (_Radius[i] + D)*(_Radius[i] + D);
    radius_core2[i] = _Radius[i] * _Radius[i] * rcore2;
    radius_surf2[i] = _Radius[i] * _Radius[i] * rsurf2;
  }

  for( int i = 0 ; i < ng3 ; i++ ) {
    nearesta[i] = -99;
  }

  return;
}
