//============================================================================//
//
//  Software Name : MEGADOCK Version 2.1
//
//  Class Name : Docking
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "docking.h"

//============================================================================//
void Docking::initialize()
//============================================================================//
{
  const int	istep = _parameter->_Cal_step;

  // FFTProcess
  _fft_process	= new FFTProcess(_cputime,_parallel,_parameter);

  if( istep != 1 ) {
    create_list();
  }

  maxsize_voxel();

  return;
}

//============================================================================//
void Docking::create_list()
//============================================================================//
{
  vector< vector<int> >	fft_table;
  const int	istep	= _parameter->_Cal_step;
  const int	nr	= _parameter->_Num_rec;
  const int	nl	= _parameter->_Num_lig;
  const int	myid	= _parallel->myid();

  fft_table.resize(nl);

  for( int i = 0 ; i < nl ; i++ ) {
    fft_table[i].resize(nr);
  }

  for( int i = 0 ; i < nl ; i++ ) {
    for( int j = 0 ; j < nr ; j++ ) {
      if( istep == 0 ) {
	if( _receptor[j].num_grid() > _ligand[i].num_grid() ) {
	  fft_table[i][j] = _receptor[j].num_grid() * 2;
	}
	else {
	  fft_table[i][j] = _ligand[i].num_grid() * 2;
	}
      }
      else if( istep == 2 ) {
	if( _receptor[j].num_grid() > _header[i].num_grid() ) {
	  fft_table[i][j] = -1;		// Not calculate this pair
	  continue;
	}
	else {
	  fft_table[i][j] = _header[i].num_grid() * 2;
	}
      }

      int k;

      for( k = 0 ; k < _FFT_list.size() ; k++ ) {
	if( _FFT_list[k] == fft_table[i][j] ) {
	  break;
	}
      }

      assert( k <= _FFT_list.size() );
      if( k == _FFT_list.size() ) {
	_FFT_list.push_back(fft_table[i][j]);
      }
    }
  }

  if( istep == 2 && myid == 0 ) {
    cerr << endl;
    for( int i = 0 ; i < nr ; i++ ) {
      cerr << "[MESSAGE] Receptor(" << _parameter->_RecPDB_file[i]
	   << ") FFT N = " << _receptor[i].num_grid()*2 << endl;
    }
    cerr << endl;

    for( int i = 0 ; i < nl ; i++ ) {
      int ligmax = 0;

      for( int j = 0 ; j < nr ; j++ ) {
	if( _receptor[j].num_grid() > ligmax ) {
	  ligmax = _receptor[j].num_grid();
	}
      }

      if( ligmax > _header[i].num_grid() ) {
	cerr << "[WARNING] The size of " << ligmax*2
	     << " is necessary for ligand(" << _parameter->_LigFFT_file[i]
	     << ") FFT N(=" << _header[i].num_grid()*2 << ").\n";
      }
    }
  }

  sort(_FFT_list.begin(),_FFT_list.end(),greater<int>());
  _Rec_list.resize(_FFT_list.size());
  _Lig_list.resize(_FFT_list.size());

  for( int i = 0 ; i < _FFT_list.size() ; i++ ) {
    for( int j = 0 ; j < nl ; j++ ) {
      for( int k = 0 ; k < nr ; k++ ) {
	if( fft_table[j][k] == _FFT_list[i] ) {
	  int	l;
	  for( l = 0 ; l < _Rec_list[i].size() ; l++ ) {
	    if( _Rec_list[i][l] == k ) {
	      break;
	    }
	  }
	    
	  assert( l <= _Rec_list[i].size() );
	  if( l == _Rec_list[i].size() ) {
	    _Rec_list[i].push_back(k);
	  }

	  for( l = 0 ; l < _Lig_list[i].size() ; l++ ) {
	    if( _Lig_list[i][l] == j ) {
	      break;
	    }
	  }
	    
	  assert( l <= _Lig_list[i].size() );
	  if( l == _Lig_list[i].size() ) {
	    _Lig_list[i].push_back(j);
	  }
	}
      }
    }
  }

  return;
}

//============================================================================//
void Docking::maxsize_voxel()
//============================================================================//
{
  int	maxfft   = 1;
  int	maxlist  = 1;
  int	maxatom  = 1;
  int	maxarray = 1;
  const int	istep = _parameter->_Cal_step;
  const int	myid = _parallel->myid();

  if( istep != 1 ) {
    for( int i = 0 ; i < _FFT_list.size() ; i++ ) {
      if( _FFT_list[i] > maxfft ) {
	maxfft = _FFT_list[i];
      }
      if( _Rec_list[i].size() > maxlist ) {
	maxlist = _Rec_list[i].size();
      }

      int n3rec = _FFT_list[i]*_FFT_list[i]*_FFT_list[i]*_Rec_list[i].size();

      if( n3rec > maxarray ) {
	maxarray = n3rec;
      }
    }

    for( int i = 0 ; i < _receptor.size() ; i++ ) {
      if( _receptor[i].num_atoms() > maxatom ) {
	maxatom = _receptor[i].num_atoms();
      }
    }
  }
  else {
    maxlist = 1;

    for( int i = 0 ; i < _parameter->_Num_lig ; i++ ) {
      const int	nligfft = _ligand[i].num_grid() * 2;

      if( nligfft > maxfft ) {
	maxfft = nligfft;
      }

      int nfft3 = nligfft * nligfft * nligfft;

      if( nfft3 > maxarray ) {
	maxarray = nfft3;
      }
    }
  }

  if( istep != 2 ) {
    for( int i = 0 ; i < _ligand.size() ; i++ ) {
      if( _ligand[i].num_atoms() > maxatom ) {
	maxatom = _ligand[i].num_atoms();
      }
    }
  }

  const int	num_grid = maxfft / 2;
  const int	nag = maxatom * num_grid;
  const int	ng3 = num_grid * num_grid * num_grid;

  if( myid == 0 ) {
    cout << "\nMaximum FFT N               = " << maxfft << endl;
    cout << "Maximum listed receptors    = " << maxlist << endl;
    cout << "Maximum number of atoms     = " << maxatom << endl;
    cout << "Maximum receptor array size = " << maxarray << endl;

    const int 	ns 	= _parameter->_Num_sort;
    const int 	nz 	= _parameter->_NZangle;
    const int 	nf3 	= maxfft * maxfft * maxfft;
    const int	nproc	= _parallel->nproc();
    const int	nproc2	= _parallel->nproc2();
    float	fmem;
    float	imem;
    
    if( istep == 0 ) {
      fmem = nproc2*(ng3*3+maxatom*3+nag*3)
	+nproc2*2*nf3+2*maxarray
	+nproc2*4*nf3
	+maxlist*(((nz-1)/nproc)*ns+ns);
      imem = nproc2*(ng3*2+maxatom*4)
	+4*maxlist*(((nz-1)/nproc)*ns+ns);
    }
    else {
      fmem = nproc2*(ng3*3+maxatom*3+nag*3)
	+2*nf3+2*maxarray
	+nproc2*4*nf3
	+maxlist*((nz-1)*ns+ns);
      imem = nproc2*(ng3*2+maxatom*4)
	+4*maxlist*((nz-1)*ns+ns);
    }

    double	mb = (fmem*4+imem*2)/1024/1024;
    mb += 10.0;		// margine
    printf("Memory requirement          = %.1f MB\n",mb);
  }

  _fft_process->alloc_array(maxlist,maxfft,maxarray);
  alloc_array(maxatom,num_grid,nag,ng3);

  return;
}

//============================================================================//
void Docking::alloc_array(const int &maxatom,const int &num_grid,const int &nag,
			  const int &ng3)
//============================================================================//
{
  const int	nproc2	  = _parallel->nproc2();

  _Grid_coord = new float[num_grid];
  if( !_Grid_coord ) {
    cerr << "[ERROR] Out of memory. Number of voxels for 1 axis = ("
	 << num_grid
	 << ") for (_Grid_coord) in docking.cpp!!\n";
    exit(1);
  }

  _Mol_coord = new float*[nproc2];
  for( int i = 0 ; i < nproc2 ; i++ ) {
    _Mol_coord[i] = new float[maxatom*3];
    if( !_Mol_coord[i] ) {
      cerr << "[ERROR] Out of memory. Number of atoms times 3 axis = ("
	   << maxatom*3
	   << ") for (_Mol_coord) in docking.cpp!!\n";
      exit(1);
    }
  }

  _Memfw = ng3*3+maxatom*3+nag*3;
  _Fwork = new float[_Memfw*nproc2];
  if( !_Fwork ) {
    cerr << "[ERROR] Out of memory. Number of float arrays = ("
	 << _Memfw*nproc2
	 << ") for (_Fwork) in docking.cpp!!\n";
    exit(1);
  }

  _Memiw = ng3*2+maxatom*4;
#ifdef PARALLEL
  const int 	no 	= _parameter->_Num_output;
  if( _Memiw < 12*no ) _Memiw = 12*no;
#endif

  _Iwork = new int[_Memiw*nproc2];
  if( !_Iwork ) {
    cerr << "[ERROR] Out of memory. Number of int arrays = ("
	 << _Memiw*nproc2
	 << ") for (_Iwork) in docking.cpp!!\n";
    exit(1);
  }

  return;
}

//============================================================================//
void Docking::rec_init(const int &ilist)
//============================================================================//
{
  const int	nfft	  = _FFT_list[ilist];
  const int	num_grid  = _FFT_list[ilist] / 2;
  const int	nlist_rec = _Rec_list[ilist].size();
  const int	ng3 	  = num_grid * num_grid * num_grid;

  assert( nlist_rec > 0 );

  _fft_process->num_fft(nfft);

  float	search_length	= _parameter->_Grid_size * num_grid;

  for( int i = 0 ; i < num_grid ; i++ ) {
    _Grid_coord[i] = search_length * ( -0.5 + ( 0.5 + i )/num_grid );
  }

  for( int i = 0 ; i < nlist_rec ; i++ ) {
    const int irec = _Rec_list[ilist][i];
    _receptor[irec].shift_center(_Mol_coord[0]);
    create_voxel(_receptor[irec],num_grid,0);
    _fft_process->receptor_fft(i,_Fwork,&_Fwork[ng3]);
  }

  return;
}

//============================================================================//
void Docking::create_voxel(Protein &rprotein,const int &num_grid,int myid2)
//============================================================================//
{
  time_t	ct1, ct2;

  if( myid2 == 0 ) {
    time(&ct1);
  }

  if( _parameter->_Score_func == 1 ) {
    rprotein.katchalski(_parameter->_Score_func,num_grid,_Grid_coord,
			_Mol_coord[myid2],&_Fwork[_Memfw*myid2]);
  }
  else if( _parameter->_Score_func == 2 ) {
    rprotein.gsc(num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		 &_Fwork[_Memfw*myid2]);
  }
  else if( _parameter->_Score_func == 3 ) {
    rprotein.psc(num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		 &_Fwork[_Memfw*myid2]);
  }
  else if( _parameter->_Score_func == 4 ) {	// FTDock electrostatic
    const float	beta = -6.0;			// MEGADOCK parameter
    rprotein.katchalski(_parameter->_Score_func,num_grid,_Grid_coord,
			_Mol_coord[myid2],&_Fwork[_Memfw*myid2]);
    rprotein.electro(beta,_parameter->_Elec_ratio,num_grid,_Grid_coord,
		     _Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		     &_Fwork[_Memfw*myid2], _parameter->grid_size());
  }
  else if( _parameter->_Score_func == 5 ) {
    rprotein.rpsc(num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		  &_Fwork[_Memfw*myid2]);
  }
  else if( _parameter->_Score_func == 6 ) {	// FTDock electrostatic
    const float	beta = -6.0;			// MEGADOCK parameter
    rprotein.rpsc(num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		  &_Fwork[_Memfw*myid2]);
    rprotein.electro(beta,_parameter->_Elec_ratio,num_grid,_Grid_coord,
		     _Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		     &_Fwork[_Memfw*myid2], _parameter->grid_size());
  }
  else if( _parameter->_Score_func == 7 ) {	// CHARMM19 electrostatic
    const float	beta = -2000.0;			// MEGADOCK parameter
    rprotein.rpsc(num_grid,_Grid_coord,_Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		  &_Fwork[_Memfw*myid2]);
    rprotein.electro(beta,_parameter->_Elec_ratio,num_grid,_Grid_coord,
		     _Mol_coord[myid2],&_Iwork[_Memiw*myid2],
		     &_Fwork[_Memfw*myid2], _parameter->grid_size());
  }

  if( myid2 == 0 ) {
    time(&ct2);
    _cputime->protein_lig_voxelization += difftime(ct2,ct1);
  }

  return;
}

//============================================================================//
void Docking::dockz(const int &ilist,const int &ilist_lig)
//============================================================================//
{
  time_t 	ct1, ct2;
  float		theta[3];

  const int	num_grid  = _FFT_list[ilist] / 2;
  const int	ng3 	  = num_grid * num_grid * num_grid;
  const int	nlist_rec = _Rec_list[ilist].size();
  const int	ilig	  = _Lig_list[ilist][ilist_lig];
  const int 	nlist_lig = _Lig_list[ilist].size();
  const int 	nc	  = _parameter->_NZangle / 10;
  const int	nproc	  = _parallel->nproc();
  const int	myid 	  = _parallel->myid();

  _fft_process->top_score_clean(nlist_rec);

  if( myid == 0 ) {
    cout << "\nLigand = " << _parameter->_LigPDB_file[ilig]
	 << " (" << ilist_lig+1 << "/" << nlist_lig << ")" << endl;
    cout << "Target receptors:" << endl;;

    for( int i = 0 ; i < nlist_rec ; i++ ) {
      cout << " " << _parameter->_RecPDB_file[_Rec_list[ilist][i]] << endl;
    }
    cout << endl;
    cout.flush();
  }

  int 	j, myid2;

#pragma omp parallel private(myid2,j,theta)
  {
#ifdef _OPENMP
    myid2 = omp_get_thread_num();
#else
    myid2 = 0;
#endif

#pragma omp for
    for( int i = myid ; i < _parameter->_NZangle ; i+=nproc ) {
      if( !((i+1)%nc) ) {
	cout << "   >Ligand rotation = " << i+1 << " / "
	     << _parameter->_NZangle << " (" << myid << "," << myid2 << ")\n";
	cout.flush();
      }

      _fft_process->rotation_index(myid2,i);
      
      for( j = 0 ; j < 3 ; j++ ) {
	theta[j] = _parameter->_Zangle[3*i+j];
      }

      ligand_rotationz(ilig,theta,myid2);
      create_voxel(_ligand[ilig],num_grid,myid2);
      _fft_process->ligand_fft(&_Fwork[_Memfw*myid2],&_Fwork[_Memfw*myid2+ng3],
      			       myid2);

      for( j = 0 ; j < nlist_rec ; j++ ) {
	_fft_process->fft_score(j,myid2);
	_fft_process->score_sort(j,myid2);
      }
    }
  }

  time(&ct1);
  _fft_process->sort_index(nlist_rec,_Fwork,_Iwork);
  time(&ct2);
  _cputime->fftprocess_sort_index += difftime(ct2,ct1);

  return;
}

//============================================================================//
void Docking::dockz1(const int &ilig)
//============================================================================//
{
  time_t 	ct1, ct2;
  float		theta[3], coord[3];
  const int	num_grid = _ligand[ilig].num_grid();
  const int	nfft 	 = num_grid * 2;
  const int	ng3 	 = num_grid * num_grid * num_grid;

  _fft_process->num_fft(nfft);

  float	search_length	= _parameter->_Grid_size * num_grid;

  for( int i = 0 ; i < num_grid ; i++ ) {
    _Grid_coord[i] = search_length * ( -0.5 + ( 0.5 + i )/num_grid );
  }

  for( int i = 0 ; i < 3 ; i++ ) {
    coord[i] = _ligand[ilig].center(i);
  }

#ifdef CHARACTER
  char		buff[256];
  ofstream 	fftlig(_parameter->_LigFFT_file[ilig].c_str());

  sprintf(buff,"%d\t%.2f\t%d",nfft,_parameter->_Grid_size,_parameter->_NZangle);
  fftlig << buff << endl;

  sprintf(buff,"\t%f\t%f\t%f",coord[0],coord[1],coord[2]);
  fftlig << _parameter->_LigPDB_file[ilig] << buff << endl;
#else
  ofstream fftlig(_parameter->_LigFFT_file[ilig].c_str(),ios_base::binary);

  fftlig.write((char*)&nfft,sizeof(int));
  fftlig.write((char*)&_parameter->_Grid_size,sizeof(float));
  fftlig.write((char*)&_parameter->_NZangle,sizeof(int));

  int	csize = sizeof(char) * (_parameter->_LigPDB_file[ilig].size()+1);

  fftlig.write((char*)&csize,sizeof(int));
  fftlig.write(_parameter->_LigPDB_file[ilig].c_str(),sizeof(char)*csize);
  fftlig.write((char*)coord,sizeof(float)*3);
#endif

  for( int i = 0 ; i < _parameter->_NZangle ; i++ ) {
    _fft_process->rotation_index(0,i);

    for( int j = 0 ; j < 3 ; j++ ) {
      theta[j] = _parameter->_Zangle[3*i+j];
    }

    ligand_rotationz(ilig,theta,0);

#ifdef CHARACTER
    sprintf(buff,"%f\t%f\t%f",theta[0],theta[1],theta[2]);
    fftlig << buff << endl;
#else
    fftlig.write((char*)theta,sizeof(float)*3);
#endif

    create_voxel(_ligand[ilig],num_grid,0);
    _fft_process->ligand_fft(_Fwork,&_Fwork[ng3],0);

    time(&ct1);
    _fft_process->fftoutl(fftlig);
    time(&ct2);
    _cputime->fftprocess_fftoutl += difftime(ct2,ct1);
  }

  fftlig.close();

  return;
}

//============================================================================//
void Docking::dock2(const int &ilist,const int &ilist_lig)
//============================================================================//
{
  time_t 	ct1, ct2;
  string	lfile;
  int		nfft, nangle, csize;
  float		bsize;
  float		coord[3], theta[3];
  char		cfile[256];
  ifstream	fftligin;
  const int	num_grid  = _FFT_list[ilist] / 2;
  const int	nlist_rec = _Rec_list[ilist].size();
  const int	ilig	  = _Lig_list[ilist][ilist_lig];
  const int 	nlist_lig = _Lig_list[ilist].size();
  const int	nproc2	  = _parallel->nproc2();
  const int	myid 	  = _parallel->myid();

  _fft_process->top_score_clean(nlist_rec);

  cout << "\nLigand = " << _parameter->_LigPDB_file[ilig]
       << " (" << ilist_lig+1 << "/" << nlist_lig << ")\n";
  cout << "Docking receptors list:\n";

  for( int i = 0 ; i < nlist_rec ; i++ ) {
    cout << _parameter->_RecPDB_file[_Rec_list[ilist][i]] << endl;
  }

#ifdef CHARACTER
  fftligin.open(_parameter->_LigFFT_file[ilig].c_str(),ios::in);
  if( !fftligin ) {
    cerr << "[ERROR]FFT file [" << _parameter->_LigFFT_file[ilig]
	 << "] not open!!" << endl;
    exit(1);
  }

  fftligin >> nfft >> bsize >> nangle;
  fftligin >> lfile >> coord[0] >> coord[1] >> coord[2];
  csize = lfile.size() + 1;
  strncpy((char *)lfile.c_str(),cfile,csize);
#else
  fftligin.open(_parameter->_LigFFT_file[ilig].c_str());
  if( !fftligin ) {
    cerr << "[ERROR]FFT file [" << _parameter->_LigFFT_file[ilig]
	 << "] not open!!" << endl;
    exit(1);
  }

  fftligin.read((char*)&nfft,sizeof(int));
  fftligin.read((char*)&bsize,sizeof(float));
  fftligin.read((char*)&nangle,sizeof(int));

  fftligin.read((char*)&csize,sizeof(int));
  fftligin.read(cfile,csize);

  lfile = string(cfile);
  fftligin.read((char*)coord,sizeof(float)*3);
#endif

  _fft_process->num_fft(nfft);
  const int nc = _parameter->_NZangle / 10;

  int i, j, im;

#pragma omp parallel
  {
    //    if( nproc2 > 1 ) _fft_process->alloc_fft(nfft);

#pragma omp for schedule(static,1) private(j,im) ordered
    for( i = 0 ; i < _parameter->_NZangle ; i++ ) {
#ifdef _OPENMP
      im = omp_get_thread_num();
#else
      im = 0;
#endif

#pragma omp ordered
      {
	if( !((i+1)%nc) ) {
	  cout << "Now ligand rotation = " << i+1 << " / "
	       << _parameter->_NZangle << " (" << myid << "," << im << ")\n";
	}

	_fft_process->rotation_index(im,i);

#ifdef CHARACTER
	fftligin >> theta[0] >> theta[1] >> theta[2];
#else
	fftligin.read((char*)theta,sizeof(float)*3);
#endif

	_parameter->_Zangle[3*i]   = theta[0];
	_parameter->_Zangle[3*i+1] = theta[1];
	_parameter->_Zangle[3*i+2] = theta[2];

	_fft_process->fftinl(fftligin,im);
      }

      for( j = 0 ; j < nlist_rec ; j++ ) {
	_fft_process->fft_score(j,im);
	_fft_process->score_sort(j,im);
      }
    }

    //    if( nproc2 > 1 ) _fft_process->delete_fft();
  }

  fftligin.close();

  time(&ct1);
  _fft_process->sort_index(nlist_rec,_Fwork,_Iwork);
  time(&ct2);
  _cputime->fftprocess_sort_index += difftime(ct2,ct1);

  return;
}

//============================================================================//
void Docking::ligand_rotationz(int ilig,float *theta,int myid2)
//============================================================================//
{
  const int	nl = _ligand[ilig].num_atoms();
  float		x, y, z;
  float		r11, r21, r31, r12, r22, r32, r13, r23, r33;

  r11 = cos(theta[0])*cos(theta[2])  -  sin(theta[0])*cos(theta[1])*sin(theta[2]);
  r21 = sin(theta[0])*cos(theta[2])  +  cos(theta[0])*cos(theta[1])*sin(theta[2]);
  r31 = sin(theta[1])*sin(theta[2]);

  r12 = -cos(theta[0])*sin(theta[2])  -  sin(theta[0])*cos(theta[1])*cos(theta[2]);
  r22 = -sin(theta[0])*sin(theta[2])  +  cos(theta[0])*cos(theta[1])*cos(theta[2]);
  r32 = sin(theta[1])*cos(theta[2]);

  r13 = sin(theta[0])*sin(theta[1]);
  r23 = -cos(theta[0])*sin(theta[1]);
  r33 = cos(theta[1]);

  for( int i = 0 ; i < nl ; i++ ) {
    x = _ligand[ilig].coordinate(i,0) - _ligand[ilig]._Center[0];
    y = _ligand[ilig].coordinate(i,1) - _ligand[ilig]._Center[1];
    z = _ligand[ilig].coordinate(i,2) - _ligand[ilig]._Center[2];
    _Mol_coord[myid2][3*i  ] = r11 * x + r12 * y + r13 * z;
    _Mol_coord[myid2][3*i+1] = r21 * x + r22 * y + r23 * z;
    _Mol_coord[myid2][3*i+2] = r31 * x + r32 * y + r33 * z;
    /*
    _ligand[ilig]._Coord[3*i  ] = r11 * x + r12 * y + r13 * z;
    _ligand[ilig]._Coord[3*i+1] = r21 * x + r22 * y + r23 * z;
    _ligand[ilig]._Coord[3*i+2] = r31 * x + r32 * y + r33 * z;
    */
  }

  return;
}

//============================================================================//
void Docking::output(const int &ilist,int ilist_lig)
//============================================================================//
{
  char		buff[256];
  float		coord[3], coordl[3], theta[3];
  int		nproc, myid;

  const int	istep 	= _parameter->_Cal_step;
  const int	nfft 	= _fft_process->num_fft();
  const float	bin 	= _parameter->_Grid_size;
  const int	no 	= _parameter->_Num_output;
  const int	nlist_rec = _Rec_list[ilist].size();
  const int	ilig	= _Lig_list[ilist][ilist_lig];

  if( istep == 2 ) {
    nproc = 1;
    myid  = 0;

    for( int i = 0 ; i < 3 ; i++ ) {
      coordl[i] = _header[ilig].center(i);
    }
  }
  else {
    nproc = _parallel->nproc();
    myid  = _parallel->myid();

    for( int i = 0 ; i < 3 ; i++ ) {
      coordl[i] = _ligand[ilig].center(i);
    }
  }

  for( int i = myid ; i < nlist_rec ; i+=nproc ) {
    const int	nr 	= _Rec_list[ilist][i];
    const int	ifile 	= _parameter->_Num_lig * nr + ilig;

    ofstream Output(_parameter->_RLOut_file[ifile].c_str());
    if( !Output ) {
      cerr << "[ERROR] Output file [" << _parameter->_RLOut_file[ifile] << "] not open!!" << endl;
      exit(1);
    }

    sprintf(buff,"%d\t%.2f",nfft,bin);
    Output << buff << endl;

    for( int j = 0 ; j < 3 ; j++ ) {
      theta[j] = _receptor[nr].angle(j);
    }

    sprintf(buff,"\t%f\t%f\t%f",theta[0],theta[1],theta[2]);
    Output << buff << endl;

    for( int j = 0 ; j < 3 ; j++ ) {
      coord[j] = _receptor[nr].center(j);
    }

    sprintf(buff,"\t%f\t%f\t%f",coord[0],coord[1],coord[2]);

    Output << _parameter->_RecPDB_file[nr] << buff << endl;

    sprintf(buff,"\t%f\t%f\t%f",coordl[0],coordl[1],coordl[2]);
    Output << _parameter->_LigPDB_file[ilig] << buff << endl;

    for( int j = 0 ; j < no ; j++ ) {
      const int n = _fft_process->top_index(i,j,0);

      for( int k = 0 ; k < 3 ; k++ ) {
	theta[k] = _parameter->_Zangle[3*n+k];

	if( theta[k] >= PI ) {
	  theta[k] -= PI*2.0;
	}
      }

      sprintf(buff,"%f\t%f\t%f\t%d\t%d\t%d\t%.2f",theta[0],theta[1],theta[2],
	      _fft_process->top_index(i,j,1),_fft_process->top_index(i,j,2),
	      _fft_process->top_index(i,j,3),_fft_process->top_score(i,j));
      Output << buff << endl;
    }

    Output.close();
  }

  return;
}
