//============================================================================//
//
//  Software Name : MEGADOCK Version 2.1
//
//  Class Name : Protein
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "protein.h"

//============================================================================//
void Protein::initialize(Parameter *rparameter,const int &ip,const int &nproc,
			 const int &myid)
//============================================================================//
{
  if( (ip%nproc) == myid ) {
    pdb_read();
  }

#ifdef PARALLEL
  pdb_comm(ip,nproc,myid);
#endif

  axis_edge();
  const int na = num_atoms();

  _Coord = new float[na*3];

  // for radius
  if( rparameter->_Score_func != 1 && rparameter->_Score_func != 4 ) {
    _Radius = new float[na];

    for( int i = 0 ; i < na ; i++ ) {
      _Radius[i] = rparameter->atom_radius(atom_type(i));
    }
  }

  // for charge
  if( rparameter->_Score_func == 7 ) {
    _Charge = new float[na];

    for( int i = 0 ; i < na ; i++ ) {
      _Charge[i] = rparameter->atom_charge(atom_type(i));
    }
  }
  else if( rparameter->_Score_func == 4 || rparameter->_Score_func == 6 ) {
    _Charge = new float[na];

    for( int i = 0 ; i < na ; i++ ) {
      _Charge[i] = rparameter->atom_ftcharge(atom_type(i));
    }

    for( int i = 0 ; i < na ; i++ ) {
      if( atom_type(i).substr(0,2) == "N " ) {
	_Charge[i] = 1.0;
	break;
      }

      if( i > 3 ) break;
    }
  }

  return;
}

//============================================================================//
void Protein::axis_edge()
//============================================================================//
{
  float coord[3];

  for( int i = 0 ; i < 3 ; i++ ) {
    _Edge[i][0] = BIG;
    _Edge[i][1] = -BIG;
  }

  for( int i = 0 ; i < num_atoms() ; i++ ) {
    for( int j = 0 ; j < 3 ; j++ ) {
      coord[j] = coordinate(i,j);
    }
#ifdef DEBUG2
    cout << "i = " << i << " X = " << coord[0] << " Y = " << coord[1]
	 << " Z = " << coord[2] << endl;
#endif
    for( int j = 0 ; j < 3 ; j++ ) {
      if( coord[j] < _Edge[j][0] ) {
	_Edge[j][0] = coord[j];
      }
      if( coord[j] > _Edge[j][1] ) {
	_Edge[j][1] = coord[j];
      }
    }
  }

  for( int i = 0 ; i < 3 ; i++ ) {
    _Center[i] = (_Edge[i][0] + _Edge[i][1]) / 2.0;
  }

  for( int i = 0 ; i < 3 ; i++ ) {
    _Angle[i] = 0.0;
  }

#ifdef DEBUG2
  for( int i = 0 ; i < 3 ; i++ ) {
    cout << "_Edge[" << i << "][0] = " << _Edge[i][0] << endl;
    cout << "_Edge[" << i << "][1] = " << _Edge[i][1] << endl;
  }
#endif

  return;
}

//============================================================================//
void Protein::shift_center(float *mol_coord)
//============================================================================//
{
  float		coord[3];
  const int	na = num_atoms();

  for( int i = 0 ; i < 3 ; i++ ) {
    coord[i] = ( _Edge[i][0] + _Edge[i][1] ) / 2.0;
  }

  // for mpicxx ?
  for( int i = 0 ; i < na ; i++ ) {
    for( int j = 0 ; j < 3 ; j++ ) {
      mol_coord[i*3+j] = coordinate(i,j) - coord[j];
      //      _Coord[i*3+j] = coordinate(i,j) - coord[j];
    }
  }

  /*
  for( int i = 0, k = 0 ; i < na ; i++ ) {
    for( int j = 0 ; j < 3 ; j++ ) {
      _Coord[k++] = coordinate(i,j) - coord[j];
    }
  }
  */

  return;
}

//============================================================================//
void Protein::katchalski_base(const int &num_grid,const float &mol_inside,
			      float *grid_coord,float *mol_coord,float *fwork)
//============================================================================//
{
  const float	surface 	= 1.0;
  const float	rjudge	 	= 1.8;
  const float	tsurface 	= 2.0;

  float		*grid_r, *grid_i;
  float		*distance3;
  float		*xd, *yd, *zd;

  const int	na = num_atoms();
  const int	nag = num_atoms() * num_grid;
  const int	ng2 = num_grid * num_grid;
  const int	ng3 = ng2 * num_grid;

  grid_r	= fwork;
  grid_i	= &fwork[ng3];
  distance3     = &fwork[ng3*2+na*3];
  xd		= &fwork[ng3*3+na*3];
  yd		= &fwork[ng3*3+na*3+nag];
  zd		= &fwork[ng3*3+na*3+nag*2];

  voxel_init(num_grid,grid_coord,mol_coord,grid_r,grid_i,distance3,xd,yd,zd);

  const float	rcore2 = rjudge * rjudge;
  const float	rtjudge = (rjudge + tsurface)*(rjudge + tsurface);

  for( int l = 0 ; l < na ; l++ ) {
    const int lc = num_grid * l;

    for( int i = 0 ; i < num_grid ; i++ ) {
      if( xd[lc+i] > rtjudge ) continue;
      const int ib = ng2 * i;

      for( int j = 0 ; j < num_grid ; j++ ) {
	if( yd[lc+j] > rtjudge ) continue;
	const float	d2 = xd[lc+i] + yd[lc+j];
	const int	jb = ib + num_grid * j;

	for( int k = 0 ; k < num_grid ; k++ ) {
	  const float d3 = d2 + zd[lc+k];
	  const int kb = jb + k;

	  if( d3 < distance3[kb] ) {
	    distance3[kb] = d3;
	  }
	}
      }
    }
  }

  for( int i = 0 ; i < ng3 ; i++ ) {
    if( distance3[i] < rtjudge ) {
      if( distance3[i] < rcore2 ) {
	grid_r[i] = mol_inside;
      }
      else {
	grid_r[i] = surface;
      }
    }
  }

#ifdef DEBUG
  cout << "Rjudge = " << rjudge << " Tsurface = " << tsurface << endl;
  cout << "Ligand grid index grid_r[i][j][k]" << endl;
  for( int i = 0 ; i < num_grid ; i++ ) {
    for( int j = 0 ; j < num_grid ; j++ ) {
      cout << "i = " << i << " j = " << j << endl;
      for( int k = 0 ; k < num_grid ; k++ ) {
	cout << grid_r[i][j][k] << " ";
      }
      cout << endl;
    }
  }
#endif

  return;
}

//============================================================================//
void Protein::voxel_init(const int &num_grid,float *grid_coord,float *mol_coord,
			 float *grid_r,float *grid_i,float *distance3,float *xd,
			 float *yd,float *zd)
//============================================================================//
{
  const int	na  = num_atoms();
  const int	nag = num_atoms() * num_grid;
  const int	ng3 = num_grid * num_grid * num_grid;

  for( int i = 0 ; i < ng3 ; i++ ) {
    grid_r[i] = 0.0;
    grid_i[i] = 0.0;
    distance3[i] = BIG;
  }

  for( int l = 0, k = 0 ; l < na ; l++ ) {
    const int l3 = l * 3;

    for( int i = 0 ; i < num_grid ; i++ ) {
      /*
      xd[k  ] = _Coord[l3  ] - grid_coord[i];
      yd[k  ] = _Coord[l3+1] - grid_coord[i];
      zd[k++] = _Coord[l3+2] - grid_coord[i];
      */
      xd[k  ] = mol_coord[l3  ] - grid_coord[i];
      yd[k  ] = mol_coord[l3+1] - grid_coord[i];
      zd[k++] = mol_coord[l3+2] - grid_coord[i];
    }
  }

  for( int i = 0 ; i < nag ; i++ ) {
    xd[i] *= xd[i];
    yd[i] *= yd[i];
    zd[i] *= zd[i];
  }

  return;
}
